#include "src/core/rom-finder.hpp"

#include <unordered_map>
#include <unordered_set>
#include <vector>
#include <thread>
#include "src/polyfill/file-search.hpp"
#include "src/polyfill/sha1.hpp"
#include "src/core/preset-controllers.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/qthread.hpp"
#include "src/core/filesystem.hpp"
#include "src/db/data-provider.hpp"
#include "src/db/transaction.hpp"

struct RomIdentifier {
	string internalName;
	uint crc32;
};

struct RomFileWithGroups {
	RomFile romFile;
	std::set<string> groups;
};

static void updateDatabase(
	std::shared_ptr<std::vector<RomFileWithGroups>> newRoms,
	std::shared_ptr<std::vector<RomFile>> changedRoms,
	std::shared_ptr<std::vector<fs::path>> lostRoms,
	std::shared_ptr<HashMap<string,RomIdentifier>> romData
) {
	SqlTransaction transaction;

	for( const fs::path &path : *lostRoms ) {
		DataProvider::deleteRomPath( path );
	}

	for( RomFileWithGroups &romAndGroups : *newRoms ) {
		RomFile &rom = romAndGroups.romFile;
		DataProvider::addRomFile( rom );
		DataProvider::addRomInfo( RomInfo{ // does nothing if the sha1 already exists
			rom.sha1,
			rom.path.stem().u8string(),
			romData->at( rom.sha1 ).internalName,
			EmulatorCore::UseDefault,
			GfxPlugin::UseDefault,
			GfxPlugin::UseDefault,
			0,
			0,
			std::move( romAndGroups.groups ),
			true,
			false,
			romData->at( rom.sha1 ).crc32,
			DefaultInputModes::Normal.id,
			false
		});
	}

	for( const RomFile &rom : *changedRoms ) {
		RomFile oldFile;
		if(
			DataProvider::tryFetchRomFile( rom.path, &oldFile ) &&
			DataProvider::countRomFilesWithSha1( oldFile.sha1 ) <= 1
		) {
			DataProvider::moveOrDeleteRomInfo( oldFile.sha1, rom.sha1 );
		} else {
			DataProvider::cloneRomInfo(
				oldFile.sha1,
				rom.sha1,
				romData->at( rom.sha1 ).internalName,
				romData->at( rom.sha1 ).crc32
			);
		}

		DataProvider::updateRomFile( rom );
	}

	transaction.commit();
}

static void recomputeRomData(
	std::shared_ptr<HashMap<fs::path,std::set<string>>> newRoms,
	std::shared_ptr<std::vector<fs::path>> changedRoms,
	std::shared_ptr<std::vector<fs::path>> lostRoms,
	const std::function<void(RomFinder::SearchResults)> &qtCallback,
	CancellationToken cancellationToken
) {
	std::shared_ptr<HashMap<string,RomIdentifier>> romData( new HashMap<string,RomIdentifier> );
	std::shared_ptr<std::vector<RomFileWithGroups>> newRomFiles( new std::vector<RomFileWithGroups> );
	std::shared_ptr<std::vector<RomFile>> changedRomFiles( new std::vector<RomFile> );

	for( auto &[path,groups] : *newRoms ) {
		if( cancellationToken.isCancelled() ) {
			QtThread::safeAsync( [qtCallback](){
				qtCallback( RomFinder::SearchResults::Cancelled );
			});
			return;
		}

		const string sha1 = Sha1::compute( path );
		newRomFiles->push_back({
			RomFile{
				path,
				RomUtil::getLastModified( path ),
				sha1
			},
			std::move( groups )
		});

		(*romData)[sha1] = {
			RomUtil::getInternalName( path ),
			RomUtil::getCrc32( path )
		};
	}

	for( const fs::path &path : *changedRoms ) {
		if( cancellationToken.isCancelled() ) {
			QtThread::safeAsync( [qtCallback](){
				qtCallback( RomFinder::SearchResults::Cancelled );
			});
			return;
		}

		const string sha1 = Sha1::compute( path );
		changedRomFiles->push_back({
			path,
			RomUtil::getLastModified( path ),
			sha1
		});

		(*romData)[sha1] = {
			RomUtil::getInternalName( path ),
			RomUtil::getCrc32( path )
		};
	}

	if( cancellationToken.isCancelled() ) return;
	QtThread::safeAsync( [=](){
		if( cancellationToken.isCancelled() ) {
			qtCallback( RomFinder::SearchResults::Cancelled );
		} else {
			updateDatabase( newRomFiles, changedRomFiles, lostRoms, romData );
			qtCallback( RomFinder::SearchResults::Updated );
		}
	});
}

static void analyzeChanges(
	std::shared_ptr<HashMap<fs::path,std::set<string>>> foundRoms,
	const std::function<void(RomFinder::SearchResults)> &qtCallback,
	CancellationToken cancellationToken
) {
	std::shared_ptr<std::vector<fs::path>> changedRoms( new std::vector<fs::path> );
	std::shared_ptr<std::vector<fs::path>> lostRoms( new std::vector<fs::path> );

	for( const RomFile &knownFile : DataProvider::fetchRomFiles() ) {
		if( foundRoms->count( knownFile.path ) > 0 ) {
			if( RomUtil::getLastModified( knownFile.path ) != knownFile.lastModified ) {
				changedRoms->push_back( knownFile.path );
			}
		} else {
			lostRoms->push_back( knownFile.path );
		}
		foundRoms->erase( knownFile.path );
	}

	if( foundRoms->empty() && changedRoms->empty() ) {
		if( lostRoms->empty() ) {
			qtCallback( RomFinder::SearchResults::NoChanges );
			return;
		}

		SqlTransaction transaction;
		for( const fs::path &path : *lostRoms ) {
			DataProvider::deleteRomPath( path );
		}
		transaction.commit();

		qtCallback( RomFinder::SearchResults::Updated );
	} else {
		std::thread( recomputeRomData, foundRoms, changedRoms, lostRoms, qtCallback, cancellationToken ).detach();
	}

}

static void scan(
	const std::function<void(RomFinder::SearchResults)> &qtCallback,
	CancellationToken cancellationToken
) {
	std::shared_ptr<HashMap<fs::path,std::set<string>>> foundRoms( new HashMap<fs::path,std::set<string>> );
	for( const RomSource &source : FileController::loadRomSources() ) {
		if( cancellationToken.isCancelled() ) {
			QtThread::safeAsync( [qtCallback](){
				qtCallback( RomFinder::SearchResults::Cancelled );
			});
			return;
		}

		const std::vector<fs::path> sourceRoms = FileSearch::find(
			source.folder,
#ifdef _WIN32
			"*.?64",
#else
			"*.[nvz]64",
#endif
			source.recursive ? source.maxDepth : (ubyte)1,
			source.followSymlinks,
			source.ignoreHidden,
			cancellationToken
		);

		for( fs::path romPath : sourceRoms ) {
#ifdef _WIN32
			const char romType = romPath.extension().u8string().c_str()[1];
			if(
				romType != 'n' && romType != 'N' &&
				romType != 'v' && romType != 'V' &&
				romType != 'z' && romType != 'Z'
			) continue;
#endif

			if( source.followSymlinks ) {
				HashSet<fs::path> visitedLinks;
				while( fs::isSymlinkSafe( romPath ) && visitedLinks.count( romPath ) == 0 ) {
					visitedLinks.insert( romPath );
					romPath = fs::read_symlink( romPath );
				}
			}

			if( fs::isRegularFileSafe( romPath ) ) {
				(*foundRoms)[romPath].insert( source.autoTags.begin(), source.autoTags.end() );
			}
		}
	}

	for( const fs::path &romPath : DataProvider::fetchManualRomPaths() ) {
		if( fs::existsSafe( romPath ) ) {
			(*foundRoms)[romPath];
		} else {
			QtThread::safeAsync( [romPath](){
				DataProvider::deleteManualRomPath( romPath );
			});
		}
	}

	QtThread::safeAsync( [=](){
		if( cancellationToken.isCancelled() ) {
			qtCallback( RomFinder::SearchResults::Cancelled );
		} else {
			analyzeChanges( foundRoms, qtCallback, cancellationToken );
		}
	});
}

CancellationToken RomFinder::scanAsync(
	const std::function<void(SearchResults)> &qtCallback
) {
	CancellationToken cancellationToken;
	std::thread( scan, qtCallback, cancellationToken ).detach();
	return cancellationToken;
}
