#include "src/ui/direct-play.hpp"

#include <QMessageBox>
#include <cstdlib>
#include "src/polyfill/sha1.hpp"
#include "src/core/preset-controllers.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/time.hpp"
#include "src/core/bps.hpp"
#include "src/core/updates.hpp"
#include "src/db/data-provider.hpp"
#include "src/ui/singleplayer-controller-select-dialog.hpp"
#include "src/ui/direct-play-window.hpp"
#include "src/ui/now-playing-window.hpp"
#include "src/ui/process-awaiter.hpp"
#include "src/ui/core-installer.hpp"
#include "src/core/retroarch.hpp"
#include "src/ui/main-window.hpp"
#include "src/input/gamepad-controller.hpp"

//TODO: remove some duplicated code from main-window.cpp

static const char *s_crashMessageParallel = ""
	"The emulator exited shortly after launching. If you are able to launch other ROMs without issues, then "
	"this ROM may contain invalid or unsafe RSP microcode that cannot be run on console accurate graphics "
	"plugins. Alternatively, if you have a very old onboard graphics card, it is possible that Vulkan is not "
	"supported on your system. In either case, using another graphics plugin might resolve the issue.";

static const char *s_crashMessageAngrylion = ""
	"The emulator exited shortly after launching. If you are able to launch other ROMs without issues, then "
	"this ROM may contain invalid or unsafe RSP microcode that cannot be run on console accurate graphics "
	"plugins. If this is the case, try running the ROM with another graphics plugin instead.";

static const char *s_crashMessageDefault = ""
	"The emulator exited shortly after launching. If you are able to launch other ROMs without issues, then "
	"this ROM is most likely corrupt.";

static inline bool hasBpsExtension( const fs::path &path ) {
	const string extension = path.extension().u8string();
	return (
		extension.length() == 4 &&
		extension[0] == '.' &&
		( extension[1] == 'b' || extension[1] == 'B' ) &&
		( extension[2] == 'p' || extension[2] == 'P' ) &&
		( extension[3] == 's' || extension[3] == 'S' )
	);
}

static inline ConnectedGamepad getActiveController( const AppSettings &settings ) {
	std::vector<ConnectedGamepad> connectedControllers = GamepadController::instance().getConnected();
	if( connectedControllers.empty() ) {
		return ConnectedGamepad{ -1, ControllerInfo() };
	} else if( connectedControllers.size() == 1 ) {
		return connectedControllers[0];
	}

	if( settings.preferredController.has_value() ) {
		for( const ConnectedGamepad &controller : connectedControllers ) {
			if( controller.info.uuid == settings.preferredController.value() ) {
				return controller;
			}
		}
	}

	SingleplayerControllerSelectDialog dialog( connectedControllers );
	dialog.exec();
	return dialog.getSelectedController();
}

static inline ControllerProfile getControllerProfile( const ConnectedGamepad &controller ) {
	if( controller.id < 0 ) {
		return DefaultProfile::XBox360;
	}

	const HashMap<Uuid,string> mappings = FileController::loadControllerMappings();
	if( mappings.count( controller.info.uuid ) > 0 ) {
		const string &activeProfile = mappings.at( controller.info.uuid );

		const std::map<string, ControllerProfile> profiles = FileController::loadControllerProfiles();
		if( profiles.count( activeProfile ) > 0 ) {
			return profiles.at( activeProfile );
		}
	}

	switch( getControllerType( controller.info.controllerId ) ) {
		case ControllerType::Gamecube:
			return DefaultProfile::Gamecube;
		case ControllerType::Nintendo64:
			return DefaultProfile::Nintendo64;
		default:
			return DefaultProfile::XBox360;
	}
}

static void playRom( const RomFile &romFile, const RomInfo &romInfo ) {
	const AppSettings &settings = FileController::loadAppSettings();

	const ConnectedGamepad activeController = getActiveController( settings );
	const EmulatorCore emulatorCore = (romInfo.emulator == EmulatorCore::UseDefault) ? settings.defaultEmulator : romInfo.emulator;
	if( !CoreInstaller::requireCore( emulatorCore ) ) return;

	GfxPlugin gfxPlugin;
	AsyncProcess *emulator = new AsyncProcess();
	try {
		*emulator = RetroArch::launchRom(
			romFile.path,
			settings,
			{{ getControllerProfile( activeController ), activeController.info.uuid }},
			romInfo,
			true,
			&gfxPlugin
		);
	} catch( ... ) {
		QMessageBox::critical( nullptr, "Emulator Missing", "Failed to launch emulator. It does not appear to be installed." );
		std::exit( 1 );
		return;
	}

	NowPlayingWindow *nowPlayingWindow = nullptr;
	if( !settings.hideWhenPlaying ) {
		nowPlayingWindow = new NowPlayingWindow( emulator, romInfo.name, romInfo.playTime );
		nowPlayingWindow->show();
	}

	ProcessAwaiter::QtSafeAwait(
		emulator,
		[=]( [[maybe_unused]] int64 exitCode, int64 runtimeMs ) {
			delete emulator;
			if( nowPlayingWindow != nullptr ) {
				nowPlayingWindow->close();
				nowPlayingWindow->deleteLater();
			}

			if( runtimeMs > 0 ) {
				DataProvider::updatePlayTime(
					romInfo.sha1,
					Time::nowMs(),
					romInfo.playTime + runtimeMs
				);
			}

			if( runtimeMs < 2000 ) {
				const char *errorMessage;
				switch( gfxPlugin ) {
					case GfxPlugin::ParaLLEl:
						errorMessage = s_crashMessageParallel;
						break;
					case GfxPlugin::Angrylion:
						errorMessage = s_crashMessageAngrylion;
						break;
					default:
						errorMessage = s_crashMessageDefault;
						break;
				}
				QMessageBox::critical( nullptr, "Possible ROM Error", errorMessage );
			}

			RetroUpdater::checkForUpdates( true, runtimeMs < 2000, false );
			std::exit( 0 );
		}
	);
}

static int runWithoutRom( QApplication &app ) {
	RetroUpdater::checkForUpdates( false, false, false );
	QGuiApplication::setQuitOnLastWindowClosed( false );
	MainWindow mainWindow;
	mainWindow.show();
	return app.exec();
}

int DirectPlay::start( QApplication &app, fs::path romPath ) {
	if( !fs::existsSafe( romPath ) ) {
		return runWithoutRom( app );
	}

	if( hasBpsExtension( romPath ) ) {
		fs::path patchedRomPath;
		Bps::BpsApplyError patchResult = Bps::tryApplyBps( romPath, patchedRomPath );
		switch( patchResult ) {
			case Bps::BpsApplyError::None:
				romPath = patchedRomPath;
				break;
			case Bps::BpsApplyError::InvalidBps:
				QMessageBox::critical( nullptr, "Patch Failed", "Failed to patch ROM. The .bps patch appears to be invalid." );
				return runWithoutRom( app );
			case Bps::BpsApplyError::PatchFailed:
				QMessageBox::critical( nullptr, "Patch Failed", "Failed to patch ROM. The base ROM does not match what the patch expects.." );
				return runWithoutRom( app );
			case Bps::BpsApplyError::NoBaseRom:
				QMessageBox::critical( nullptr, "Patch Failed", "Failed to patch ROM. The base rom required to patch is not known to Parallel Launcher." );
				return runWithoutRom( app );
			case Bps::BpsApplyError::WriteError:
				QMessageBox::critical( nullptr, "Patch Failed", "Failed to patch ROM. An error occurred while writing the patched ROM to disk." );
				return runWithoutRom( app );
		}
	}

	RomInfo romInfo;
	RomFile romFile;
	if( DataProvider::tryFetchRomByPath( romPath, false, &romInfo, &romFile ) ) {
		playRom( romFile, romInfo );
		return app.exec();
	}

	romFile = RomFile{
		romPath,
		RomUtil::getLastModified( romPath ),
		Sha1::compute( romPath )
	};


	if( DataProvider::tryFetchRomByHash( romFile.sha1, false, &romInfo ) ) {
		DataProvider::addRomFile( romFile );
		playRom( romFile, romInfo );
		return app.exec();
	}

	DirectPlayWindow mainWindow( romFile );
	QObject::connect( &mainWindow, &DirectPlayWindow::play, [&](){
		if( !RomUtil::coveredBySearchPath( romPath ) ) {
			DataProvider::addManualRomPath( romPath );
		}

		romInfo = mainWindow.getRomInfo();
		DataProvider::addRomInfo( romInfo );
		DataProvider::addRomFile( romFile );
		playRom( romFile, romInfo );
	});

	mainWindow.show();
	QRect geo = mainWindow.geometry();
	geo.setHeight( 0 );
	mainWindow.setGeometry( geo );
	return app.exec();
}
